#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <math.h>

/*--------------------------------------------------------------------*/
/*                                                                    */
/*   This programs displays some of the most important data sent over */
/*   a Motorola trunked radio system control channel.                 */
/*   Right now it expects an interface described in the above text    */
/*   to sit at COM1                                                   */
/*   If you want improvements you have to do them yourself.           */
/*                                                                    */
/*   This is not be used by anyone for any monetary gain. If you do   */
/*   we hope Motorola's lawyers will crush the life out of you.       */
/*                                                                    */
/*   Complaints? Remember - you get what you paid for. And you didn't */
/*   pay a cent for this program.                                     */
/*                                                                    */
/*   Still got Complaints? Well here are our canned responses:        */
/*                                                                    */
/*       - The code is the comment                                    */
/*       - What do you expect? I work for Microsoft                   */
/*       - Cheesiest is easiest                                       */
/*       - I'll put the comments in later                             */
/*       - If it was hard to write it should be hard to read          */
/*       - Go ahead and try to find me                                */
/*       - Rember - you always have a fiend in Jesus                  */
/*                                                                    */
/*--------------------------------------------------------------------*/
/*                                                                    */
/* port info  :  COM1                                if dlab bit =1   */
/*      0x3f8 : receive/transmit data                     baud0       */
/*      0x3f9 : interrupt enable register     - IER       baud1       */
/*      0x3fa : interrupt identifier register - IIR                   */
/*      0x3fb : line control register         - LCR                   */
/*      0x3fc : modem control register        - MCR                   */
/*      0x3fd : line status register          - LSR                   */
/*      0x3fe : modem status register         - MSR                   */
/*                                                                    */
/*--------------------------------------------------------------------*/
/*                                                                    */
/* PORT INFO  : 8259 INTERRUPT CONTROLLER                             */
/*      0X021 : A CORRESPONDING LOW BIT ENABLES A SPECIFIC INTERRUPT  */
/*              COM1 IS IRQ4, AND THEREFORE CONTROLLED BY BIT 0X10    */
/*      0X020 : WHENEVER AN INTERRUPT IS RESOLVED, A 0X20 SHOULD BE   */
/*              PORTED OUT HERE TO GENERATE AN END OF INTERRUPT       */
/*              SIGNAL.                                               */
/*                                                                    */
/*--------------------------------------------------------------------*/

/* global variables */
int lc=0;
double baudrate = 9600.0; /* baud rate of computer to scanner link */
                          /* Note: for exmaple/vestigal code purposes */
int fobp=0;	/* pointer to current position in array fob	      */
char ob[100];   /* output buffer for packet before being sent to screen */
int obp=0;	/* pointer to current position in array ob		*/
int aid[51],naid=0,uaid[51];

struct osw_stru{
  int cmd;
  int id;
  int grp;
  int ok;
};

struct osw_stru bosw[20];


static unsigned int buflen= 15000;      /* length of data buffer      */
static volatile unsigned int cpstn = 0; /* current position in buffer */
static unsigned int fdata[15001] ;      /* frequency data array       */

void interrupt (*oldfuncc) (); /* vector to old com port interrupt    */

/**********************************************************************/
/*                                this is serial com port interrupt   */
/* we assume here that it only gets called when one of the status     */
/* lines on the serial port changes (that's all you have hooked up).  */
/* All this handler does is read the system timer (which increments   */
/* every 840 nanoseconds) and stores it in the fdata array. The MSB   */
/* is set to indicate whether the status line is zero. In this way    */
/* the fdata array is continuously updated with the appropriate the   */
/* length and polarity of each data pulse for further processing by   */
/* the main program.						      */
void interrupt com1int()
{
  static unsigned int d1,d2,ltick,tick,dtick;

  /* the system timer is a 16 bit counter whose value counts down     */
  /* from 65535 to zero and repeats ad nauseum. For those who really  */
  /* care, every time the count reaches zero the system timer 	      */
  /* interrupt is called (remember that thing that gets called every  */
  /* 55 milliseconds and does housekeeping such as checking the       */
  /* keyboard.                                                        */
  outportb (0x43, 0x00);       /* latch counter until we read it      */
  d1 = inportb (0x40);         /* get low count                       */
  d2 = inportb (0x40);         /* get high count                      */

  /* get difference between current, last counter reading             */
  tick  = (d2 << 8) + d1;
  dtick = ltick - tick;
  ltick = tick;

  if ((inportb(0x3fe) & 0xF0) > 0) dtick = dtick | 0x8000;
			      else dtick = dtick & 0x3fff;

  fdata[cpstn] = dtick;        /* put freq in fdata array             */
  cpstn  ++;                   /* increment data buffer pointer       */
  if (cpstn>buflen) cpstn=0;   /* make sure cpstn doesnt leave array  */

  d1 = inportb (0x03fa);       /* clear IIR                           */
  d1 = inportb (0x03fd);       /* clear LSR                           */
  d1 = inportb (0x03fe);       /* clear MSR                           */
  d1 = inportb (0x03f8);       /* clear RX                            */
  outportb (0x20, 0x20);       /* this is the END OF INTERRUPT SIGNAL */
                               /* "... that's all folks!!!!"          */
}

void set8250(double bps)       /* sets up the 8250 UART               */
{
  static unsigned int t,dv,tp;

  dv = (int) 1843200.0/(16.0*bps);
  /* how to configure your serial port setup:          */
  /* do you want two stop bits? then make sure you set bit 2 in tp */
  /* do you want a parity bit generated? then set bit bit 3 in tp  */
  /* do you want even parity? then set bit 4 in tp                 */
  /* do you want an 8 bit word length? set bits 0 and 1 in tp      */
  /* do you want an 7 bit word length? set bit  1 in tp            */
  tp = 0x80 + 0x02 + 0x01;
  outportb (0x3fb, tp);  /* set line control register  */
  outport  (0x3f8, dv);  /* output brg divisor latch   */
  tp = tp & 0x7f;        /* switch out brg divisor reg */
  outportb (0x3fb, tp);


  /* only enable those interrupts caused by changing levels on the    */
  /* status lines                                                     */
  outportb (0x03fb, 0x00);     /*  set IER on 0x03f9                  */
  outportb (0x03f9, 0x08);     /*  enable MODEM STATUS INTERRUPT      */

  outportb (0x03fc, 0x0a);     /*  push up RTS, DOWN DTR              */

  t = inportb(0x03fd);         /*  clear LSR                          */
  t = inportb(0x03f8);         /*  clear RX                           */
  t = inportb(0x03fe);         /*  clear MSR                          */
  t = inportb(0x03fa);         /*  clear IID                          */
  t = inportb(0x03fa);         /*  clear IID - again to make sure     */
}

void set8253()                 /*  set up the 8253 timer chip         */
{                              /* NOTE: ctr zero, the one we are using*/
                               /*  is incremented every 840nSec, is   */
                               /*  main system time keeper for dos    */
  outportb (0x43, 0x34);       /* set ctr 0 to mode 2, binary         */
  outportb (0x40, 0x00);       /* this gives us the max count         */
  outportb (0x40, 0x00);
}

/****************************************************************/

void show_time()
{
  static unsigned int t;
  t = bosw[1].id;
  gotoxy(41,1); printf ("System Date/Time");
  gotoxy(41,2); printf ("Year : %02u",t>>9);
  gotoxy(41,3); printf ("Month: %02u",(t>>5)&0x0f);
  gotoxy(41,4); printf ("Day  : %02u",t&0x1f);
  t = bosw[0].id;
  gotoxy(41,5); printf ("Time : %2u:%02u",(t>>8)&0x1f,t&0xff);
}

void show_unk()
{
  gotoxy(41,10); printf ("Unknown Info");
  gotoxy(41,11); printf ("%04X  %04X",bosw[0].id,bosw[1].id);
}

void show_unk2()
{
  gotoxy(41,13); printf("Com. 32B: %04X",bosw[0].id);
}

void show_unk3()
{
  gotoxy(41,14); printf("Com. 309: %04X",bosw[0].id);
}

void show_tgr()
{
  static int oy=2;
  if (oy == 25) movetext(60,3,80,25,60,2) ; else oy++;
  gotoxy(60,oy);
  printf ("%04X --> %04X",bosw[1].id,bosw[0].id);
}

void show_freak()
{
  static int nos=0,i,st=0,x,y,te;
  static double fr;

  /* if the command could be frequency assignment keep track of it */
    st = 0;
    for (i=0; i< naid; i++)
    {
      if (aid[i] == bosw[0].cmd)
      {
	st ++;
	x = i/24;
	y = 1 + i - (x*24);
	x = 10 + (x * 20);
	gotoxy(x,y);
	if (bosw[0].grp == 1) printf ("G "); else printf("I ");
	printf("%04X",bosw[0].id);
	uaid[i]=1;
      }
    }
    if (st == 0)  /* id not found */
    {
      aid[i]=bosw[0].cmd;
      uaid[naid]=1;
      x = i / 24;
      y = 1 + i - (x*24);
      x = 1 + (x * 20);
      gotoxy(x,y);
      te = bosw[0].cmd ^ 0x004;
      fr = 851.1125 + (0.025 * (double) te);
      printf("%8.7g ",fr);
      if (bosw[0].grp == 1) printf ("G "); else printf("I ");
      printf("%04X",bosw[0].id);
      if (naid<48) naid++;
    }

  nos++;
  if (nos == 40)   /* every 40 OSW's : mark out inactive channels */
  {
    for (i=0; i<naid; i++)
    {
      if (uaid[i] == 0)
      {
	x = i / 24;
	y = 1 + i - (x*24);
	x = 10 + (x * 20);
	gotoxy(x,y);
	printf("      ");
      }
      uaid[i] = 0;
    }
    nos = 0;
  }
}

/* send a serial character. Vestigal code just for example purpose */
void send_char(char c)
{
  /* wait for transmit reg to empty */
  while ( (inportb(0x3fd) & 0x20) == 0x00);

  /* send character */
  outportb(0x3f8,c);
}

/* THIS routine is called every time a new OSW has been received    */
/* This would be the place to put a routine that controls a scanner */
/* right now it tries displays selected OSW data to the screen.     */
/* Trunked Frequency assignements are shown along with the active   */
/* Talkgroups. */
void show_osw()
{

  if ((bosw[0].cmd <  0x2d0) && (bosw[1].cmd != 0x308)) show_freak();
  if ((bosw[0].cmd == 0x322) && (bosw[1].cmd == 0x308)) show_time();
  if ((bosw[1].cmd == 0x308) && (bosw[0].cmd == 0x310)) show_tgr();
  if ((bosw[1].cmd == 0x308) && (bosw[0].cmd == 0x30b)) show_unk();
  if (bosw[0].cmd == 0x32b) show_unk2();
  if (bosw[0].cmd == 0x309) show_unk3();

  /* let's say you at this point wanted to tell your scanner to go   */
  /* to some frequency and you have it hooked up to the RS-232 port. */
  /* Send a series of send_char commands giving the right command    */
  /* codes for your scanner.                                         */
  /* The baud rate is defined in the global variables section.       */
  /* The setup is in the set8250 routine                             */
  /* Example:
  send_char(69); send_char(65); send_char(84); send_char(32);
  send_char(77); send_char(69);
  */

}

/* process de-interleaved incoming data stream */
/* inputs: -1 resets this routine in prepartion for a new OSW       */
/*         48 zero bit                                              */
/*	   49 one bit                                               */
/* once an OSW is received it will be placed into the buffer bosw[] */
/* bosw[0] is the most recent entry				    */
void proc_osw(int sl)
{
  static int ct,l,sr,sax,f1,f2,iid,cmd,neb=0;
  static char osw[50],gob[100];


  if (sl == -1) ct = 0;
  else
  {
    gob[ct] = sl;
    ct++;

      if (ct == 76)
      {
	sr = 0x036E;
	sax = 0x0393;
	neb = 0;

	/* run through convolutional error correction routine */
	/* Reference : US PATENT # 4055832		      */
	for (l=0; l<76; l+=2)
	{
	  osw[l>>1] = gob[l];   /* osw gets the DATA bits   */
	  if (gob[l] == 49)     /* gob becomes the syndrome */
	  {
	    gob[l  ] ^= 0x01;
	    gob[l+1] ^= 0x01;
	    gob[l+3] ^= 0x01;
	  }
	}
	for (l=0; l<76; l+=2)   /* now correct errors       */
	{
	  if ( (gob[l+1] == 49) && (gob[l+3] == 49) )
	  {
	    osw[l >> 1] ^= 0x01;
	    gob[l+1] ^= 0x01;
	    gob[l+3] ^= 0x01;
	  }
	}

	/* run through error detection routine */
	for (l=0; l<27; l++)
	{
	  if ((sr & 0x01) != 0 ) sr = (sr >> 1) ^ 0x0225; else sr = sr >> 1;
	  if (osw[l] == 49) sax = sax ^ sr;
	}
	for (l=0; l<10; l++)
	{
	  if (osw[36-l] == 49) f1 = 0; else f1 = 1;
	  if ((sax & 0x01) > 0) f2 = 1; else f2 = 0;
	  sax = sax >> 1;
	  if (f1 != f2) neb++;  /* neb counts # of wrong bits */
	}

	/* slide down bosw buffer */
	for (l=10; l>=1; l--) bosw[l]=bosw[l-1];

	/* if no errors - OSW received properly; process it */
	if (neb == 0)
	{
	  for (l=0; l<16; l++)
	  {
	    iid = iid << 1;
	    if (osw[l] == 48) iid++;
	  }
	  iid = iid ^ 0x33C7;
	  bosw[0].ok = 1;
	  bosw[0].id = iid;
	  if (osw[16] == 48) bosw[0].grp = 1; else bosw[0].grp = 0;
	  cmd = 0;
	  for (l = 17; l<27; l++)
	  {
	    cmd = cmd << 1;
	    if (osw[l] == 48) cmd++;
	  }
	  cmd = cmd ^ 0x032A;
	  bosw[0].cmd = cmd;

	  show_osw();

	}
	else bosw[0].ok = 0;
      }
  }
}

void pigout()  /* de-interleave OSW stored in array ob[] and process */
{
  static int i1,i2,k;
    proc_osw(-1);
    for (i1=0; i1<19; i1++)
    {
      for (i2=0; i2<4; i2++)
      {
	k = (i2*19) + i1;
	proc_osw( (int) ob[k]);
      }
    }
}

/* this routine looks for the frame sync and splits off the 76 bit */
/* data frame and sends it on for further processing               */
void frame_sync(char gin)
{
  static int sr = 0, fs = 0xAC, bs=85;

  /* keep up 8 bit sliding register for sync checking purposes */
  sr = sr << 1;
  sr = sr & 0xff;
  if (gin == 49) sr = sr | 0x01;

  ob[bs-1] = gin;

  /* if sync seq and enough bits are found - data block */
  if ((sr == fs) && (bs > 83))
  {
    pigout();  /* the latest frame is stored in array ob[] */
    bs = 0;
  }
  if (bs < 98) bs++;
}

void main (int argc)
{
  unsigned int n,i=0,cw1=49,cw0=48;
  char s=48;
  double dt,exc=0.0,clk=0.0,xct;

  if (argc > 1)
  {
    cw1 = 48;
    cw0 = 49;
  }

  clrscr();

  gotoxy(60,1);
  printf ("ID   --> TALK GROUP");

  /* dt is the number of expected clock ticks per bit */
  dt =  1.0/(3600.0*838.8e-9);

  oldfuncc = getvect(0x0c);    /* save COM1 Vector                    */
  setvect (0x0c, com1int);     /* Capture COM1 vector                 */


  n = inportb (0x21);          /* enable IRQ4 interrupt               */
  outportb(0x21, n & 0xef);

  set8253();                   /* set up 8253 timer chip              */

  set8250(baudrate);           /* set up 8250 UART                    */

  while (kbhit() == 0)
  {
   if (i != cpstn)
   {
    if  ( ( fdata[i] & 0x8000) != 0) s = cw0; else s = cw1;

    /* add in new number of cycles to clock  */
    clk += (fdata[i] & 0x7fff);
    xct = exc + 0.5 * dt;  /* exc is current boundary */
    while ( clk >= xct )
    {
      frame_sync(s);   /* send raw bit stream to first processing routin */
      clk = clk - dt;
    }
    /* clk now holds new boundary position. update exc slowly... */
    /* 0.005 sucks; 0.02 better; 0.06 mayber even better; 0.05 seems pretty good */
    exc = exc + 0.0250*(clk - exc);

    i++;
    if( i >buflen) i = 0;
   }

  }

  outportb (0x21, n);          /* disable IRQ4 interrupt              */
  setvect (0x0c, oldfuncc);    /* restore old COM1 Vector             */

  gotoxy(1,24);

}


